""" __init__.py

This module implements the Gaussian elimination of a matrix by column
reductions.  Coefficients are assumed to lie on a finite field Z (mod p) Where
p is a prime and Z are the integers.  Ideally, this should be implemented in
Cython or similar later on.
"""
